#include "vfl.h"
#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <stdarg.h>

static char *code;
static int pc;
static int sp;
static int csp;
static int lsp;
static int32_t *stack;
static uint32_t *callStack;
static uint32_t *loopStack;
static int32_t *mem;
static int memSize;
static int stackSize;

#if ERROR_HANDLING
typedef struct {
	int line;
	int col;
} CodePos;

static int curCommandPos = 0;

static CodePos getLineAndCol(int pos) {
	CodePos codePos;
	codePos.line = 1;
	codePos.col = 0;
	
	for (int i = 0; i <= pos; i++) {
		if (code[i] == '\n') {
			codePos.line++;
			codePos.col = 0;
		} else if (i != pos)
			codePos.col++;
	}
	
	return codePos;
}

static void runtimeError(char *fmt, ...) {
	CodePos pos = getLineAndCol(curCommandPos);
	fprintf(stderr,
		"line %d, col %d: ",
		pos.line, pos.col
	);
	
	va_list args;
	va_start(args, fmt);
	vfprintf(stderr, fmt, args);
	va_end(args);
	
	exit(1);
}
#endif

static int isDigit(char c) {
	return c >= '0' && c <= '9';
}

static int isLetter(char c) {
	return c >= 'a' && c <= 'z';
}

static void memory_init(int l_memSize, int l_stackSize) {
	memSize = l_memSize;
	stackSize = l_stackSize;
	
	mem = calloc(memSize, sizeof(int32_t));
	stack = calloc(stackSize, sizeof(int32_t));
	callStack = calloc(stackSize, sizeof(int32_t));
	loopStack = calloc(stackSize, sizeof(int32_t));
	
	if (
		mem == NULL
		|| stack == NULL
		|| callStack == NULL
		|| loopStack == NULL
	) {
		fprintf(stderr, "failed to allocate memory\n");
		exit(1);
	}
}

static void memory_free() {
	free(mem);
	free(stack);
	free(callStack);
	free(loopStack);
}

static inline void memory_set(uint32_t addr, int32_t value) {
	#if ERROR_HANDLING
	if (addr < 0 || addr >= memSize) {
		runtimeError("out of bounds memory write (address %d)\n", addr);
	}
	#endif
	mem[addr] = value;
}

static inline int32_t memory_get(uint32_t addr) {
	#if ERROR_HANDLING
	if (addr < 0 || addr >= memSize) {
		runtimeError("out of bounds memory read (address %d)\n", addr);
	}
	#endif
	return mem[addr];
}

static inline void push(int32_t value) {
	#if ERROR_HANDLING
	if (sp >= stackSize) {
		runtimeError("stack overflow\n");
	}
	#endif
	stack[sp++] = value;
}

static inline int32_t pop() {
	#if ERROR_HANDLING
	if (sp <= 0) {
		runtimeError("stack underflow\n");
	}
	#endif
	return stack[--sp];
}

static inline int32_t peek(int n) {
	#if ERROR_HANDLING
	if (n < 0) {
		runtimeError("stack underflow\n");
	}
	if (n >= sp) {
		runtimeError("stack overflow\n");
	}
	#endif
	return stack[n];
}

static inline void cpush(int32_t value) {
	#if ERROR_HANDLING
	if (csp >= stackSize) {
		runtimeError("call stack overflow");
	}
	#endif
	callStack[csp++] = value;
}

static inline int32_t cpop() {
	#if ERROR_HANDLING
	if (csp <= 0) {
		runtimeError("unmatched '}'\n");
	}
	#endif
	return callStack[--csp];
}

static inline void lpush(int32_t value) {
	#if ERROR_HANDLING
	if (lsp >= stackSize) {
		runtimeError("loop stack overflow\n");
	}
	#endif
	loopStack[lsp++] = value;
}

void vfl_init(int memSize, int stackSize) {
	memory_free();
	memory_init(memSize, stackSize);
	
	sp = 0;
	csp = 0;
	lsp = 0;
}

void vfl_run(char *l_code) {
	code = l_code;
	pc = 0;
	int32_t a, b, c;
	
	while (1) {
		#if ERROR_HANDLING
		curCommandPos = pc;
		#endif
		
		char op = code[pc];
		pc++;
		if (op == 0) {
			#if ERROR_HANDLING
			if (lsp > 0) {
				runtimeError("unmatched '['\n");
			}
			#endif
			break;
		}
		
		if (isDigit(op)) {
			int32_t d = (op - '0');
			while (isDigit(code[pc])) {
				d = d * 10 + (code[pc] - '0');
				pc++;
			}
			push(d);
		
		} else if (isLetter(op)) {
			push(op - 'a');
		
		} else switch (op) {
			case '`':
				while (1) {
					#if ERROR_HANDLING
					if (code[pc] == 0) {
						runtimeError("unmatched '`'\n");
					}
					#endif
					if (code[pc] == '`') {
						pc++;
						break;
					}
					pc++;
				}
				break;
			
			case '\'': push(code[pc++]); break;
			case '$': a = pop(); push(a); push(a); break;
			case '_': a = pop(); break;
			case '\\': b = pop(); a = pop(); push(b); push(a); break;
			case '@': c = pop(); b = pop(); a = pop(); push(b); push(c); push(a); break;
			case '?': a = pop(); push(peek(sp-1-a)); break;
			case ':': b = pop(); a = pop(); memory_set(b, a); break;
			case ';': a = pop(); push(memory_get(a)); break;
			case '+': b = pop(); a = pop(); push(a+b); break;
			case '-': b = pop(); a = pop(); push(a-b); break;
			case '*': b = pop(); a = pop(); push(a*b); break;
			case '/':
				b = pop(); a = pop();
				#if ERROR_HANDLING
				if (b == 0) runtimeError("division by 0\n");
				#endif
				push(a/b);
				break;
			case '%':
				b = pop(); a = pop();
				#if ERROR_HANDLING
				if (b == 0) runtimeError("modulo by 0\n");
				#endif
				push(a%b);
				break;
			case '&': b = pop(); a = pop(); push(a&b); break;
			case '|': b = pop(); a = pop(); push(a|b); break;
			case '~': a = pop(); push(~a); break;
			case '=': b = pop(); a = pop(); push(a==b ? ~0 : 0); break;
			case '>': b = pop(); a = pop(); push(a>b ? ~0 : 0); break;
			case '<': b = pop(); a = pop(); push(a<b ? ~0 : 0); break;
			case '!': a = pop(); cpush(pc); pc = a; break;
			case '.': b = pop(); a = pop(); portOut(b, a); break;
			case ',': a = pop(); push(portIn(a)); break;
			
			case '"':
				a = pop();
				while (1) {
					#if ERROR_HANDLING
					if (code[pc] == 0) {
						runtimeError("unmatched '\"'\n");
					}
					#endif
					if (code[pc] == '"') {
						pc++;
						break;
					} else if (code[pc] == '\\') pc++;
					
					portOut(a, code[pc]);
					pc++;
				}
				break;
			
			case '{':
				push(pc);
				a = 1;
				while (1) {
					#if ERROR_HANDLING
					if (code[pc] == 0) {
						runtimeError("unmatched '{'\n");
					}
					#endif
					if (code[pc] == '{')
						a++;
					else if (code[pc] == '}') {
						a--;
						if (a == 0) {
							pc++;
							break;
						}
					}
					pc++;
				}
				break;
			case '}': pc = cpop(); break;
			
			case '(':
				a = pop();
				if (a) break;
				
				a = 1;
				while (1) {
					#if ERROR_HANDLING
					if (code[pc] == 0) {
						runtimeError("unmatched '('\n");
					}
					#endif
					if (code[pc] == '(')
						a++;
					else if (code[pc] == ')') {
						a--;
						if (a == 0) {
							pc++;
							break;
						}
					}
					pc++;
				}
				break;
			case ')': break;
			
			case '[': lpush(pc); break;
			case ']': case '#':
				#if ERROR_HANDLING
				if (lsp <= 0) {
					runtimeError(
						op == ']'
						? "unmatched ']'\n"
						: "'#' used outside of loop\n"
					);
				}
				#endif
				pc = loopStack[lsp-1];
				break;
			
			case '^':
				#if ERROR_HANDLING
				if (lsp <= 0) {
					runtimeError("'^' used outside of loop\n");
				}
				#endif
				lsp--;
				a = 1;
				while (code[pc] != 0) {
					if (code[pc] == '[')
						a++;
					else if (code[pc] == ']') {
						a--;
						if (a == 0) {
							pc++;
							break;
						}
					}
					pc++;
				}
				break;
		}
	}
}
